# XPyBind
#
# Copyright (C) 2004 Dan Aloni <da-x@colinux.org>
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
# 
# ratpoison is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA

import os

# Exception handling

class Error(Exception): pass
class FocusLostError(Error) : pass


# KeySymMod is a tuple which contains a KeySym and a canoical representation
# of the modifiers. 

def _keysymmod_to_string(keysym, modifiers):
    """
    Convert between a keysym tuple (for example, (110, 4)), to a string
    representation of that tuple (C-t, for example).
    """
    mods = []
    for val, mod in xpb_modifiers:
        if modifiers & val:
            mods.append(mod)

    return '-'.join([[],['-'.join(mods)]][len(mods) != 0] + [xpybind_c.keysym_to_string(keysym)])

def _string_to_keysymmod(s):
    """
    Convert between a string representation of a key combination (C-t, for example),
    to a keysym tuple (for example, (110, 4)).
    """
    parts = s.split('-')
    mod_parts = parts[:-1]
    key = parts[len(parts)-1]
    mods = []
    modifiers = 0

    for val, mod in xpb_modifiers:
        if mod in mod_parts:
            modifiers |= val
	    
    return (xpybind_c.string_to_keysym(key), modifiers)

def _grab_key(s):
    """
    Tells X to send as an event when the user press the specified key combination,
    which is given as a string ("C-t", for example). event_key will be called with
    the key combination when it is pressed.
    """
    keysymmod = _string_to_keysymmod(s)
    xpybind_c.grab_key(*keysymmod)

def _ungrab_key(s):
    """
    Tells X to stop from sending as an event when the user press the specified
    key combination, which is given as a string ("C-t", for example).
    """
    
    keysymmod = _string_to_keysymmod(s)
    xpybind_c.ungrab_key(*keysymmod)

def _read_key():
    """
    Blocks until the user hits a key combination. The key combination is returned
    as a string ('C-t', for example). If the focus from xpybind is lost while the
    function waits, FocusLostError is raised.
    """
    
    keysymmod = xpybind_c.read_key()
    if keysymmod == "focus lost":
        raise FocusLostError()
    return _keysymmod_to_string(*keysymmod)

def exit():
    """
    Tells the main loop of xpybind to exit.
    """
    xpybind_c.exit()

def notify_user(text, timeout=0):
    """
    This function displays a non-intrusive notification for the user. It is
    used by xpybind to notify the user about the currently expected hotkey
    when key sequences are entered. The Window displays itself without taking
    the keyboard focus from the user.

    The user can also use this function inside ~/.xpybind for notification
    about the success of operations that were carried out by key sequences
    or other activities.

    @timeout - The number of seconds until the window disappears from the
    moment it pops.
    """
    xpybind_c.uodate_input_window(text, timeout)
    
def _event_key(keysym, modifiers):
    """
    This function is called by xpybind.c when one of the keys that were bound
    by grab_key() was hit. The two parameters that are passed to this function
    aer the KeySym and the canonical representation of the modifiers.

    By examining the keyboard_maps tree, the function decides whether to:
    1. Wait for more keys by using read_key()
    2. Call a function that is bound to the key combination sequence according
       to the tree.
    3. Do nothing in the case where nothing is bound to the key combination
       sequence it intercepted.
    """

    s = _keysymmod_to_string(keysym, modifiers)
    keys_got = []
    keys_got.append(s)

    print s
    if s not in keyboard_maps:
        return
    
    last_window = xpybind_c.focus_input_window()
    xpybind_c.create_input_window()

    current_keyboard_map_node = keyboard_maps

    try:
        unmatched = False
        unmatched_key = None
        while True:
            if s in current_keyboard_map_node:
                (type, data) = current_keyboard_map_node[s]
                if type == 'sub':
                    current_keyboard_map_node = data
                    print "Expecting from list: %s" % ', '.join(current_keyboard_map_node.keys())
                    notify_user(' %s {%s}' % (
                        ' '.join(keys_got),
                        ','.join(current_keyboard_map_node.keys())))
                    s = _read_key()
                    keys_got.append(s)
                    print "Got %s" % s
                elif type == 'func':
                    data()
                    break
            else:
                unmatched = True
                unmatched_key = s
                break
        xpybind_c.unfocus_input_window(*last_window)

        if not unmatched:
            xpybind_c.destroy_input_window()
        else:
            notify_user(' unmatched %s' % (unmatched_key, ), 500)

    except FocusLostError:
        print "Focus lost"
        xpybind_c.unfocus_input_window(0)
        xpybind_c.destroy_input_window()
    
def _map_keys():
    """
    Take the keyboard_maps table from user settings and create the
    keyboard_maps tree from it. Also bind all the keys that are in the
    first level of the tree so we would be able to intercept all the
    key combination sequences that the tree represents.
    """
    for sequence, func in user_config.keyboard_maps:
        current_keyboard_map_node = keyboard_maps
        parts = sequence.split(" ")
        for part in parts[:-1]:
            (type, data) = current_keyboard_map_node.setdefault(part, ('sub', {}))
            if type != 'sub':
                break
            
            current_keyboard_map_node = data

        last_part = parts[len(parts)-1]
        current_keyboard_map_node[last_part] = ('func', func)

    for key in keyboard_maps:
        _grab_key(key)

    print "Grabbing %s" % ' '.join(keyboard_maps.keys())

def _unmap_keys():
    """
    Unmap the keys that were mapped by map_keys().
    """
    print "Ungrabbing %s" % ' '.join(keyboard_maps.keys())
    
    for key in keyboard_maps:
        _ungrab_key(key)
        
def _get_user_config():
    """
    Load user settings. If something fails, then use a default,
    quite empty keyboard_maps table.    
    """
    import imp
    user_config = imp.new_module('xpybind_user_config')
    print "Loading settings from %s" % (user_settings_filename, )
    execfile(user_settings_filename, vars(user_config), vars(user_config))
    return user_config

def _load_user_config():
    global user_config
    try:
        config = _get_user_config()
    except:
	from traceback import print_exc
        print "Error loading user settings from %s" % (user_settings_filename, )
	print_exc()
        return
    user_config = config

def reload_user_config():
    """
    Reload the user's configuration. It is useful to bind this function
    to a key combination sequence in the user settings so the user would
    be able to reload her settings.
    """

    _unmap_keys()
    _load_user_config()
    _map_keys()
    
def _default_user_config():
    global user_config
    import imp

    user_config = imp.new_module('xpybind_user_config')
    user_config.keyboard_maps = [
        ("Menu r", reload_user_config),
    ]

def _init(settings_filename=None):
    """
    Called by xpybind.c to initialize the module.
    """

    global keyboard_maps
    keyboard_maps = {}

    global xpybind_c
    import xpybind_c
    
    global xpb_modifiers
    xpb_modifiers = zip(*xpybind_c.get_xpb_modifiers())

    print "xpybind python extension initialized"

    global user_settings_filename
    if settings_filename:
        user_settings_filename = settings_filename
    else:
        user_settings_filename = "%s/.xpybind" % os.getenv("HOME")

    _default_user_config()
    _load_user_config()
    _map_keys()
